#
# Copyright 2018 Antoine Sanner
# 
# ### MIT license
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import pytest

import numpy as np
from NuMPI.Tools import  Reduction
from runtests.mpi import MPITest
from tests.MPI_minimization_problems import MPI_Objective_Interface
import scipy.optimize
import  time
import tests.minimization_problems as mp

from NuMPI.Optimization.MPI_LBFGS_Matrix_H import steepest_descent_wolfe2, LBFGS

def timer(fun, *args, **kwargs):
    start = time.perf_counter()
    res = fun(*args, **kwargs)
    delay = time.perf_counter() - start
    return res, delay

def test_linesearch():
    pass


def test_MPI_Parallel_Interface(comm): # TODO: Duplicate with test_LBFGS_interface ?
    """ Tests the parallel wrapping of objectives
    Test if parallel Version gives the same as the serial Version
    :param comm:
    :return:
    """

    def printMPI(msg):
        for i in range(comm.Get_size()):
            comm.barrier()
            if comm.Get_rank() == i:
                print("Proc {}: {}".format(i, msg))
    n = 10

    par = MPI_Objective_Interface(mp.Extended_Rosenbrock,nb_domain_grid_pts=n,comm=comm)

    printMPI(par.counts)

    ref = mp.Extended_Rosenbrock

    np.testing.assert_array_equal(mp.Extended_Rosenbrock.startpoint(n)[par.subdomain_slices], par.startpoint())
    np.testing.assert_almost_equal(mp.Extended_Rosenbrock.f(mp.Extended_Rosenbrock.startpoint(n)),par.f(par.startpoint()),err_msg="Different Function Value at startpoint")
    np.testing.assert_allclose(mp.Extended_Rosenbrock.grad(mp.Extended_Rosenbrock.startpoint(n))[par.subdomain_slices],
                                   par.grad(par.startpoint()), err_msg="Different Gradient Value at startpoint")


#@pytest.mark.parametrize("PObjective",[mp.Extended_Rosenbrock]) # Objective should support parallelization !
#@pytest.mark.parametrize("n",[10,20])

def test_analytical_min(comm):
    """
    Compares the result with the analyticaly known posistion of the minimum
    :return:
    """

    def printMPI(msg):
        for i in range(comm.Get_size()):
            comm.barrier()
            if comm.Get_rank() == i:
                print("Proc {}: {}".format(i, msg))

    n = 50

    Objective = mp.Extended_Rosenbrock

    PObjective=MPI_Objective_Interface(Objective, nb_domain_grid_pts=n, comm=comm)

    x0 = PObjective.startpoint()

    res = LBFGS(PObjective.f_grad, x0, jac=True, maxcor=5, maxiter=100,
                gtol=1e-12, ftol=0, pnp= Reduction(comm))
    #                        ^ only terminates if gradient condition is satisfied
    assert res.success
    assert res.message=="CONVERGENCE: NORM_OF_GRADIENT_<=_GTOL"
    np.testing.assert_allclose(res.x,PObjective.xmin(),atol=1e-16,rtol = 1e-7)

    assert np.abs(res.fun-Objective.minVal(n))< 1e-7


def test_ftol(comm):
    pass

def test_gtol(comm):
    pass

def test_g2tol(comm):
    pass

def test_alltol(comm):
    pass

@pytest.mark.skip(reason="just plotting")
def test_time_complexity(comm):


    maxcor = 5
    Objective = mp.Extended_Rosenbrock
    n= np.array([10,100,1000,1e4,1e5,1e6],dtype = int)
    t = np.zeros(len(n),dtype=float)
    res = [None] * len(n)
    pnp = Reduction(comm)
    for i in range(len(n)):
        PObjective = MPI_Objective_Interface(Objective, nb_domain_grid_pts=n[i], comm=comm)
        x0 = PObjective.startpoint()

        res[i],t[i] = timer(LBFGS,PObjective.f_grad, x0, jac=True, maxcor=maxcor, maxiter=100000, gtol=(1e-5), pnp=pnp)

        assert res[i].success


    if True:
        import matplotlib.pyplot as plt

        fig,ax = plt.subplots()
        ax.plot(n,t/n,'+-', label = "time / DOF")
        ax.plot(n, [t[i] / n[i] / res[i].nit for i in range(len(n))], '+-',label = "time per DOF per iteration")
        ax2 = plt.twinx(ax)
        ax2.plot(n,[res[i].nit for i in range(len(n))],'o',label = "nit")
        ax.set_xscale('log')
        ax.set_yscale('log')
        ax.set_ylabel("t/n (s)")
        ax.set_xlabel("DOF")
        ax.legend()
        ax2.legend()
        #ax.plot(n,n,c='gray')
        plt.show(block = True)



#
# Copyright 2018 Antoine Sanner
# 
# ### MIT license
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
serial test
"""

import numpy as np
import scipy.optimize
from NuMPI.Optimization.LBFGS_Matrix_H import LBFGS
import tests.minimization_problems as mp

import pytest
@pytest.mark.parametrize("Objective",[mp.Extended_Rosenbrock]) # Only these where Solution is analytically known
@pytest.mark.parametrize("n",[10,20])
def test_analytical_min(Objective,n):
    """
    Compares the result with the analyticaly known posistion of the minimum
    :return:
    """

    x0 = Objective.startpoint(n)

    res = LBFGS(Objective.f, x0, jac=Objective.grad, maxcor=5, gtol=1e-6, maxiter=1000)

    #np.testing.assert_almost_equal( res.x,Objective.xmin(n))
    np.testing.assert_allclose(np.reshape(res.x, (-1,)), np.reshape(Objective.xmin(n),(-1,)), rtol=1e-7)

@pytest.mark.parametrize("Objective",[mp.Trigonometric])
@pytest.mark.parametrize("n",[10,30])
def test_compare_scipy(Objective,n):
    x0 = Objective.startpoint(n)

    resLBGFS = LBFGS(Objective.f, x0, jac=Objective.grad, maxcor=5, gtol=1e-8, maxiter=1000)
    assert resLBGFS.success
    #Patch
    #resScipy = scipy.optimize.minimize(lambda x: Objective.f(np.reshape(x,(-1,1))),np.reshape(x0,(-1,)), jac = lambda x : np.reshape(Objective.grad(np.reshape(x,(-1,1))),(-1,)),method='L-BFGS-B',tol = 1e-4)
    resScipy = scipy.optimize.minimize(Objective.f, np.reshape(x0, (-1,)),
                                      jac= Objective.grad,method="L-BFGS-B",
                                      options=dict(gtol = 1e-8))
    assert resScipy.success

    np.testing.assert_allclose( np.reshape(resLBGFS.x,(-1,)),resScipy.x, rtol= 1e-2, atol = 1e-5)  #TODO: The location of the minimumseems o be hard to find
    assert np.abs(Objective.f(resScipy.x) - Objective.f(resLBGFS.x)) < 1e-8


#
# Copyright 2019 Lars Pastewka
#           2019 Antoine Sanner
# 
# ### MIT license
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


"""
MPI-parallel writing of matrices in numpy's 'npy' format.
"""

import numpy as np
from .. import MPI
import numpy as np
import struct
import os.path
import abc

from numpy.compat import asstr
from numpy.lib.format import magic, MAGIC_PREFIX, _filter_header
from numpy.lib.utils import safe_eval


def save_npy(fn, data, subdomain_locations=None, nb_grid_pts=None, comm=MPI.COMM_WORLD):
    """

    Parameters
    ----------
    data : numpy array : data owned by the processor
    location : index of the first element of data within the global data
    nb_grid_pts : nb_grid_pts of the global data
    comm : MPI communicator

    Returns
    -------

    """
    data = np.asarray(data)
    ndims = len(data.shape)



    if ndims==1:
        data = data.reshape((-1,1))

    if subdomain_locations is None:
        subdomain_locations = (0,0)
    elif ndims ==1:
        subdomain_locations = (subdomain_locations, 0)
    nb_subdomain_grid_pts = data.shape

    if nb_grid_pts is None:
        nb_grid_pts = nb_subdomain_grid_pts
    elif ndims ==1:
        nb_grid_pts = (nb_grid_pts, 1)

    fortran_order = np.isfortran(data)

    from numpy.lib.format import dtype_to_descr, magic
    magic_str = magic(1, 0)
    arr_dict_str = str({'descr': dtype_to_descr(data.dtype),
                        'fortran_order': fortran_order,
                        'shape': (nb_grid_pts[0],) if ndims==1 else nb_grid_pts})

    while (len(arr_dict_str) + len(magic_str) + 2) % 16 != 15:
        arr_dict_str += ' '
    arr_dict_str += '\n'
    header_len = len(arr_dict_str) + len(magic_str) + 2

    file = MPI.File.Open(comm, fn, MPI.MODE_CREATE | MPI.MODE_WRONLY)
    if comm.Get_rank() == 0:
        file.Write(magic_str)
        file.Write(np.int16(len(arr_dict_str)))
        file.Write(arr_dict_str.encode('latin-1'))

    if fortran_order:
        # the returned array will be in fortran_order.
        # the data is loaded in C_contiguous array but in a transposed manner
        # data.transpose() is called which swaps the shapes back again and
        # toggles C-order to F-order
        ix = 1
        iy = 0
    else:
        ix = 0
        iy = 1

    mpitype = MPI._typedict[data.dtype.char]
    filetype = mpitype.Create_vector(nb_subdomain_grid_pts[ix],
                                     # number of blocks  : length of data in the non-contiguous direction
                                     nb_subdomain_grid_pts[iy],  # length of block : length of data in contiguous direction
                                     nb_grid_pts[iy]
                                     # stepsize: the data is contiguous in y direction,
                                     # two matrix elements with same x position are separated by ny in memory
                                     )  # create a type
    # see MPI_TYPE_VECTOR

    filetype.Commit()  # verification if type is OK
    file.Set_view(
        header_len + (subdomain_locations[ix] * nb_grid_pts[iy]
                      + subdomain_locations[iy]) * mpitype.Get_size(),
        filetype=filetype)
    if fortran_order:
        data = data.transpose()
    file.Write_all(data.copy())  # TODO: is the copy needed ?
    filetype.Free()

    file.Close()

class MPIFileTypeError(Exception):
    pass


class MPIFileIncompatibleResolutionError(Exception):
    pass


def mpi_read_bytes(file, nbytes):
    # allocate the buffer
    buf = np.empty(nbytes, dtype=np.int8)
    file.Read_all(buf)
    return buf.tobytes()


def load_npy(fn, subdomain_locations=None, nb_subdomain_grid_pts=None, comm=MPI.COMM_WORLD):
    file = MPIFileViewNPY(fn, comm)
    #if file.nb_grid_pts != nb_domain_grid_pts:
    #    raise MPIFileIncompatibleResolutionError(
    #        "nb_domain_grid_pts is {} but file nb_grid_pts is {}".format(nb_domain_grid_pts, file.nb_grid_pts))
    data = file.read(subdomain_locations, nb_subdomain_grid_pts)
    file.close()
    return data


class MPIFileView(metaclass=abc.ABCMeta):
    def __init__(self, fn, comm):
        self.fn = fn
        self.comm = comm

    @abc.abstractmethod
    def _read_header(self):
        pass

    @abc.abstractmethod
    def read(self):
        pass


def make_mpi_file_view(fn, comm, format=None):  # TODO: DISCUSS: oder als __init__ von der MPIFileView Klasse ?
    readers = {
        "npy": MPIFileViewNPY
    }

    if not os.path.isfile(fn):
        raise FileExistsError("file {} not found".format(fn))
    # TODO: chack existence of file also with parallel reader.

    if format is not None:
        try:
            reader = readers[format]
        except KeyError:
            raise (ValueError("Given format is not recognised, you should give {}".format(readers.keys())))
        return reader(fn, comm)

    for reader in readers.values():
        try:
            return reader(fn, comm)
        except MPIFileTypeError:
            pass
    raise MPIFileTypeError("No MPI filereader was able to open_topography the file {}".format(fn))

class MPIFileViewNPY(MPIFileView):
    """

    you may have a look at numpy.lib.format if you want to understand how this code works
    """

    def __init__(self, fn, comm):
        super().__init__(fn, comm)
        self._read_header()

    def detect_format(self):  # TODO: maybe useless
        try:
            self._read_header()
            return True
        except:
            return False

    def _read_header(self):
        self.file = None
        try:
            magic_str = magic(1, 0)
            self.file = MPI.File.Open(self.comm, self.fn, MPI.MODE_RDONLY)  #
            magic_str = mpi_read_bytes(self.file, len(magic_str))
            if magic_str[:-2] != MAGIC_PREFIX:
                raise MPIFileTypeError("MAGIC_PREFIX missing at the beginning of file {}".format(self.fn))

            version = magic_str[-2:]

            if version == b'\x01\x00':
                hlength_type = '<H'
            elif version == b'\x02\x00':
                hlength_type = '<I'
            else:
                raise MPIFileTypeError("Invalid version %r" % version)

            hlength_str = mpi_read_bytes(self.file, struct.calcsize(hlength_type))
            header_length = struct.unpack(hlength_type, hlength_str)[0]
            header = mpi_read_bytes(self.file, header_length)

            header = _filter_header(asstr(header))
            d = safe_eval(header)  # TODO: Copy from _read_array_header  with all the assertions
            self.dtype = np.dtype(d['descr'])
            self.fortran_order = d['fortran_order']
            self.nb_grid_pts = d['shape']
        except Exception as err:
            # FIXME! This should be handled through a resource manager
            if self.file is not None:
                self.file.Close()
            raise err

    def read(self, subdomain_locations=None, nb_subdomain_grid_pts=None):

        if subdomain_locations is None:
            subdomain_locations = (0,0)
        if nb_subdomain_grid_pts is None:
            nb_subdomain_grid_pts = self.nb_grid_pts

        # Now how to start reading ?

        mpitype = MPI._typedict[self.dtype.char]


        if self.fortran_order:
            # the returned array will be in fortran_order.
            # the data is loaded in C_contiguous array but in a transposed manner
            # data.transpose() is called which swaps the shapes back again and
            # toggles C-order to F-order
            ix = 1
            iy = 0
        else:
            ix = 0
            iy = 1

        # create a type
        filetype = mpitype.Create_vector(
            nb_subdomain_grid_pts[ix],  # number of blocks  : length of data in the non-contiguous direction
            nb_subdomain_grid_pts[iy],  # length of block : length of data in contiguous direction
            self.nb_grid_pts[iy]
            # stepsize: the data is contiguous in y direction,
            # two matrix elements with same x position are separated by ny in memory
        )

        filetype.Commit()  # verification if type is OK
        self.file.Set_view(
            self.file.Get_position() + (
                    subdomain_locations[ix] * self.nb_grid_pts[iy] + subdomain_locations[iy]) * mpitype.Get_size(),
            filetype=filetype)

        data = np.empty((nb_subdomain_grid_pts[ix], nb_subdomain_grid_pts[iy]),
                        dtype=self.dtype)

        self.file.Read_all(data)
        filetype.Free()

        if self.fortran_order:
            data = data.transpose()

        return data

    def close(self):
        self.file.Close()


#
# Copyright 2019 Lars Pastewka
#           2018-2019 Antoine Sanner
# 
# ### MIT license
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#



import numpy as np
import os
import warnings

from NuMPI.IO.MPIFileIO import save_npy, load_npy,  MPIFileIncompatibleResolutionError, MPIFileViewNPY, MPIFileTypeError
from NuMPI import MPI
import pytest

testdir = os.path.dirname(os.path.realpath(__file__))

def test_FileSave_1D(comm):
    nb_domain_grid_pts = 128
    np.random.seed(2)
    globaldata = np.random.random(nb_domain_grid_pts)

    nprocs = comm.Get_size()
    rank = comm.Get_rank()

    step = nb_domain_grid_pts // nprocs

    if rank == nprocs - 1:
        subdomain_slices = slice(rank * step, None)
        subdomain_locations =rank * step
        nb_subdomain_grid_pts = nb_domain_grid_pts - rank * step
    else:
        subdomain_slices = slice(rank * step, (rank + 1) * step)
        subdomain_locations = rank * step
        nb_subdomain_grid_pts = step

    localdata = globaldata[subdomain_slices]

    save_npy("test_Filesave_1D.npy",
             localdata,
             subdomain_locations,
             nb_domain_grid_pts,
             comm)
    comm.barrier() # The MPI_File reading and closing doesn't have to finish together
    loaded_data = np.load("test_Filesave_1D.npy")
    np.testing.assert_array_equal(loaded_data,globaldata)

    comm.barrier()
    if rank == 0:
        os.remove("test_Filesave_1D.npy")

@pytest.mark.skipif(MPI.COMM_WORLD.Get_size()> 1,
        reason="test is only serial")
def test_FileSave_1D_list():
    nb_domain_grid_pts = 8
    np.random.seed(2) 
    globaldata = np.random.random(nb_domain_grid_pts).tolist()

    save_npy("test_Filesave_1D_list.npy",globaldata)

    loaded_data = np.load("test_Filesave_1D_list.npy")
    np.testing.assert_array_equal(loaded_data,globaldata)

    os.remove("test_Filesave_1D_list.npy")

@pytest.fixture(scope="module", params=("C", "F"))
def globaldata(request, comm):
    order = request.param

    rank = comm.Get_rank()
    nprocs = comm.Get_size()

    nb_domain_grid_pts = (128,128)
    np.random.seed(2)
    if order=="C":
        globaldata = np.random.random(nb_domain_grid_pts)
        assert globaldata.flags["C_CONTIGUOUS"]
    elif order=="F":
        globaldata = np.random.random(nb_domain_grid_pts[::-1]).transpose()
        assert globaldata.flags["F_CONTIGUOUS"]
    if rank == 0:
        np.save("test_FileLoad_2D.npy", globaldata)

    comm.barrier()

    yield globaldata
    comm.barrier()
    if rank == 0:
        os.remove("test_FileLoad_2D.npy")


### Helper function:

class DistributedData:
    def __init__(self, data, nb_domain_grid_pts, subdomain_locations):
        self.data = data
        self.nb_domain_grid_pts = nb_domain_grid_pts
        self.nb_subdomain_grid_pts = data.shape
        self.subdomain_locations = subdomain_locations

    @property
    def subdomain_slices(self):
        return tuple([slice(s, s + n) for s, n in
                      zip(self.subdomain_locations, self.nb_subdomain_grid_pts)])

def make_2d_slab_y(comm, globaldata):

    nprocs = comm.Get_size()
    rank = comm.Get_rank()

    nb_domain_grid_pts = globaldata.shape

    step = nb_domain_grid_pts[1] // nprocs
    if rank == nprocs - 1:
        subdomain_slices = (slice(None, None), slice(rank * step, None))
        subdomain_locations = [0, rank * step]
        nb_subdomain_grid_pts = [nb_domain_grid_pts[0], nb_domain_grid_pts[1] - rank * step]
    else:
        subdomain_slices = (slice(None, None), slice(rank * step, (rank + 1) * step))
        subdomain_locations = [0, rank * step]
        nb_subdomain_grid_pts = [nb_domain_grid_pts[1], step]

    return DistributedData(globaldata[subdomain_slices], nb_domain_grid_pts, subdomain_locations)

def make_2d_slab_x(comm, globaldata):
    """
    returns the part of globaldata attribute to the present rank in 2D data decomposition
    Parameters
    ----------
    comm : communicator
    globaldata

    Returns
    -------
    DistributedData
    """
    nprocs = comm.Get_size()
    rank = comm.Get_rank()

    nb_domain_grid_pts = globaldata.shape

    step = nb_domain_grid_pts[0] // nprocs

    if rank == nprocs - 1:
        subdomain_slices = (slice(rank * step, None), slice(None, None))
        subdomain_locations = [rank * step, 0]
        nb_subdomain_grid_pts = [nb_domain_grid_pts[0] - rank * step, nb_domain_grid_pts[1]]
    else:
        subdomain_slices = (slice(rank * step, (rank + 1) * step), slice(None, None))
        subdomain_locations = [rank * step, 0]
        nb_subdomain_grid_pts = [step, nb_domain_grid_pts[1]]

    return DistributedData(globaldata[subdomain_slices], nb_domain_grid_pts, subdomain_locations)

@pytest.mark.parametrize("decompfun",[make_2d_slab_x, make_2d_slab_y])
def test_FileSave_2D(decompfun, comm, globaldata):
    distdata = decompfun(comm, globaldata)

    save_npy("test_Filesave_2D.npy",
             distdata.data,
             distdata.subdomain_locations,
             distdata.nb_domain_grid_pts, comm)
    comm.barrier()
    if comm.Get_rank() == 0:
        loaded_data = np.load("test_Filesave_2D.npy")
        np.testing.assert_array_equal(loaded_data, globaldata)

        os.remove("test_Filesave_2D.npy")
    comm.barrier()

@pytest.mark.parametrize("decompfun",[make_2d_slab_x, make_2d_slab_y])
def test_FileView_2D(decompfun, comm, globaldata):
    distdata = decompfun(comm, globaldata)

    #arr = np.load("test_FileLoad_2D.npy")
    #assert arr.shape == self.nb_domain_grid_pts

    file = MPIFileViewNPY("test_FileLoad_2D.npy", comm=comm)

    assert file.nb_grid_pts == distdata.nb_domain_grid_pts
    assert file.dtype == globaldata.dtype

    loaded_data = file.read(nb_subdomain_grid_pts=distdata.nb_subdomain_grid_pts,
                            subdomain_locations= distdata.subdomain_locations)


    np.testing.assert_array_equal(loaded_data, distdata.data)


@pytest.mark.parametrize("decompfun",[make_2d_slab_x, make_2d_slab_y])
def test_FileLoad_2D(decompfun, comm, globaldata):
    distdata = decompfun(comm, globaldata)

    loaded_data = load_npy("test_FileLoad_2D.npy",
                           nb_subdomain_grid_pts=distdata.nb_subdomain_grid_pts,
                           subdomain_locations=distdata.subdomain_locations,
                           comm=comm)

    np.testing.assert_array_equal(loaded_data, distdata.data)


@pytest.fixture
def npyfile():
    """
    defines a filename and makes a cleanup once the test was executed
    Yields
    ------
    filename

    """
    yield "test_same_numpy.npy"
    try:
        os.remove("test_same_numpy.npy")
    except FileNotFoundError:
        pass

@pytest.mark.skip(reason="just some statements on numpy behaviour")
def test_detect_fortran_order(comm_self):
    # fix some statements on numpy behaviour

    # Prepare fortran array
    arr = np.array(range(6), dtype=float).reshape(3, 2)
    arr = arr.transpose()

    # States numpy behaviour
    assert arr.shape == (2, 3)
    assert arr.flags["C_CONTIGUOUS"] is False
    assert arr.flags["F_CONTIGUOUS"] is True

    np.save("test.npy", arr)

    # asserts the loaded array is exactly the same
    loaded = np.load("test.npy")
    assert loaded.shape == (2, 3)
    assert loaded.flags["C_CONTIGUOUS"] is False
    assert loaded.flags["F_CONTIGUOUS"] is True


def test_load_same_numpy_load(comm_self, npyfile):
    data = np.random.random(size=(2, 3))
    np.save(npyfile, data)
    loaded_data = load_npy(npyfile, comm=comm_self)
    np.testing.assert_allclose(loaded_data, data)


def test_same_numpy_load_transposed(comm_self, npyfile):
    data = np.random.random(size=(2, 3)).T
    np.save(npyfile, data)
    loaded_data = load_npy(npyfile, comm=comm_self)
    np.testing.assert_allclose(loaded_data, data)


def test_load_same_numpy_save(comm_self,npyfile):
    data = np.random.random(size=(2, 3))
    save_npy(npyfile, data, comm=comm_self)
    loaded_data = np.load(npyfile)
    np.testing.assert_allclose(loaded_data, data)
    assert np.isfortran(data) == np.isfortran(loaded_data)


def test_same_numpy_save_transposed(comm_self, npyfile):
    data = np.random.random(size=(2, 3)).T
    save_npy(npyfile, data, comm=comm_self)
    loaded_data = np.load(npyfile)
    np.testing.assert_allclose(loaded_data, data)
    assert np.isfortran(data) == np.isfortran(loaded_data)


#@pytest.mark.filterwarnings("error: ResourceWarnings")
# unfortunately the ResourceWarnings are never raised even when the files are not closed
def test_raises_and_no_resourcewarnings(comm_self):
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter(
            "always")  # deactivate hiding of ResourceWarnings

        with pytest.raises(MPIFileTypeError):
            load_npy(os.path.join(testdir, "wrongnpyfile.npy"), comm=comm_self)

        # assert no warning is a ResourceWarning
        for wi in w:
            assert not issubclass(wi.category, ResourceWarning)

def test_corrupted_file(comm_self):
    """
    tests that the reader behaves decently when trying to open a file having
    the wrong format see issue #23
    """
    # create test corrupted file

    with open("corrupted.dummy", "w") as f:
        f.write("dfgdfghlkjhgiuhdfg")

    with pytest.raises(MPIFileTypeError):
        MPIFileViewNPY("corrupted.dummy", comm=comm_self)

def test_filestream(comm_self, npyfile):
    data = np.random.normal(size=(4,6))

    np.save(npyfile, data)
    with open(npyfile, mode="r") as f:
        read_data = load_npy(npyfile, comm=comm_self)
        np.testing.assert_allclose(read_data, data)


#
# Copyright 2018 Antoine Sanner
# 
# ### MIT license
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#



import numpy as np
import scipy.optimize
from NuMPI.Optimization.LBFGS_Matrix_H import LBFGS

from NuMPI.Optimization.Wolfe import second_wolfe_condition,first_wolfe_condition
import pytest

def test_3D():
    # Gaussian Potential

    # quadratic
    toPlot = False


    def ex_fun(x):
        "x should be an np array"
        #x.shape = (-1, 1)
        return np.sum(np.dot((x ** 2).flat,np.array([1,4,9])),axis=0)

    def ex_jac(x):
        return 2 * np.array((1* x[0],4* x[1],9 * x[2]))



    xg, yg = np.linspace(-5, 5, 51), np.linspace(-6, 6, 51)

    def mat_fun(x_g, x_):
        Z = np.zeros((xg.size, yg.size))

        for i in range(Z.shape[0]):
            for j in range(Z.shape[1]):
                Z[j, i] = ex_fun(np.array([xg[i], yg[j], 0]))
        return Z

    ## plot
    if toPlot:
        import matplotlib
        #matplotlib.use("MACOSX")
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        X, Y = np.meshgrid(xg, yg)
        ax.contour(X, Y, mat_fun(xg, yg))
        fig.show()

        print(ex_fun(np.array((0, 0, 0))))
        fig2, ax2 = plt.subplots()
        for z in [0]:
            for y in [0, 1, 2]:
                ax2.plot(xg, [ex_fun(np.array((x, y, z))) for x in xg])
                ax2.plot(xg, [ex_jac(np.array((x, y, z)))[0]for x in xg],'--')
        #plt.show(block = True)
        ######11
    # Initial history:
    x_old = np.array([2.,1.,-1.],dtype=float)
    x_old.shape = (-1,1)
    #x_old.shape=(-1,1)
    grad_old = ex_jac(x_old)

    # line search
    reslinesearch = scipy.optimize.minimize_scalar(fun=lambda alpha: ex_fun(x_old - grad_old * alpha), bounds=(0, 2000),
                                          method="bounded")
    assert reslinesearch.success
    print("alpha {}".format(reslinesearch.x))
    x = x_old - grad_old * reslinesearch.x
    assert ex_fun(x) < ex_fun(x_old)
    grad = ex_jac(x)
    print("x first_linesearch {}".format(x))
    k = 1
    print("Wolfe")

    print("1st: {}".format(first_wolfe_condition(ex_fun, x_old, ex_jac, -grad_old, reslinesearch.x, beta1=1e-4)))
    print("2nd {}".format(second_wolfe_condition(x_old,ex_jac,-grad_old,reslinesearch.x,beta2=0.9)))


    resscipy = scipy.optimize.minimize(ex_fun,x,jac = ex_jac)
    print("schipy success: {}".format(resscipy.success))

    res = LBFGS(ex_fun, x, jac=ex_jac, x_old=x_old, maxcor=3, maxiter=10)
    print("nit {}".format(res.nit))

    if toPlot:
        import matplotlib
        #matplotlib.use("MACOSX")
        import matplotlib.pyplot as plt
        fig3d = plt.figure()
        ax3d = fig3d.add_subplot(111, projection='3d')

        for it, i in zip(res['iterates'], range(len(res['iterates']))):
            ax3d.plot(it.x[0], it.x[1],it.x[2], '+k')
            #ax3d.annotate(i, it.x)
        fig3d.show()
        plt.show(block=True)

    np.testing.assert_almost_equal(res.x,np.zeros(res.x.shape),decimal = 5)

def test_quadratic_nD():


    # quadratic
    n = 50;

    factors = np.random.random(n)+0.1

    def ex_fun(x):
        "x should be an np array"
        # x.shape = (-1, 1)
        return np.sum(np.dot((x ** 2).flat, factors**2), axis=0)

    def ex_jac(x):
        return 2 * np.diag(factors).dot(x)

    ## plot

    ######11
    # Initial history:
    x_old = np.random.random(n)
    x_old.shape = (-1, 1)
    # x_old.shape=(-1,1)
    grad_old = ex_jac(x_old)

    # line search
    alpha, phi, phi0, derphi = scipy.optimize.linesearch.scalar_search_wolfe2(
        lambda alpha: ex_fun(x_old - grad_old * alpha),
        lambda alpha: np.dot(ex_jac(x_old - grad_old * alpha).T, -grad_old))
    assert derphi is not None
    x = x_old - grad_old * alpha
    assert ex_fun(x) < ex_fun(x_old)
    grad = ex_jac(x)
    print("x first_linesearch {}".format(x))
    k = 1

    resscipy = scipy.optimize.minimize(ex_fun, x, jac=ex_jac)
    print("schipy success: {}".format(resscipy.success))

    #print(x)
    #print(ex_jac(x))

    res = LBFGS(ex_fun, x, jac=ex_jac, x_old=x_old, maxcor=5, maxiter=100,g2tol=1e-10)
    print("nit {}".format(res.nit))

    if False:
        import matplotlib
        matplotlib.use("MACOSX")
        import matplotlib.pyplot as plt
        fig3d = plt.figure()
        ax3d = fig3d.add_subplot(111, projection='3d')

        for it, i in zip(res['iterates'], range(len(res['iterates']))):
            ax3d.plot(it.x[0], it.x[1], it.x[2], '+k')
            # ax3d.annotate(i, it.x)
        fig3d.show()
        plt.show(block=True)

    np.testing.assert_almost_equal(res.x, np.zeros((n,1)), decimal=4)

@pytest.mark.parametrize("n", [3,5,10,20])  # This has Problems with Linesearch at high number of points
def test_gaussian_nD(n):


    factors = np.random.random(n)+0.5
    shift   = (np.random.random(n)-0.5) * 0.1
    shift.shape = (-1,1)

    def ex_fun(x):
        "x should be an np array"
        # x.shape = (-1, 1)
        return - np.exp(- np.sum(np.dot(((x-shift) ** 2).flat, factors**2), axis=0))

    def ex_jac(x):
        return - 2 * np.diag(factors).dot(x-shift) * ex_fun(x)

    ## plot

    ######11
    # Initial history:
    x=np.random.random(n) -0.5
    x.shape = (-1, 1)


    #resscipy = scipy.optimize.minimize(ex_fun, x, jac=ex_jac)
    #print("schipy success: {}".format(resscipy.success))


    res = LBFGS(ex_fun, x, jac=ex_jac, maxcor=2, gtol=1e-5, maxiter=10000)
    print("nit {}".format(res.nit))

    if False:
        import matplotlib
        matplotlib.use("MACOSX")
        import matplotlib.pyplot as plt
        fig3d = plt.figure()
        ax3d = fig3d.add_subplot(111, projection='3d')

        for it, i in zip(res['iterates'], range(len(res['iterates']))):
            ax3d.plot(it.x[0], it.x[1], it.x[2], '+k')
            # ax3d.annotate(i, it.x)
        fig3d.show()
        plt.show(block=True)

    np.testing.assert_allclose(res.x, shift, atol = 1e-3)

def test_x2_xcosy():


    def ex_fun(x_):
        x=x_.reshape((-1,1))
        return .5 * x[0, 0] ** 2 + x[0, 0] * np.cos(x[1, 0])

    def ex_jac(x_):
        x = x_.reshape((-1, 1))
        return np.array([[x[0, 0] + np.cos(x[1, 0])],
                         [-x[0, 0] * np.sin(x[1, 0])]])
    ## plot


    #plt.show(block=True)
    ######
    # Initial history:
    x = np.array([[3], [-4]],dtype=float)

    k = 1

    res = LBFGS(ex_fun, x, jac=ex_jac, gtol = 1e-5 ,maxcor=5, maxiter=10000,linesearch_options=dict(c1 = 1e-4,c2 = 0.999))
    print("nit {}".format(res.nit))

    ref = scipy.optimize.minimize(ex_fun, x.reshape(-1), jac=lambda x: ex_jac(x).reshape(-1))

    assert ref.success

    if False:
        import matplotlib
        matplotlib.use("MACOSX")
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        xg, yg = np.linspace(-5, 5, 51), np.linspace(-6, 6, 51)

        def mat_fun(xg, yg):
            Z = np.zeros((xg.size, yg.size))
            for i in range(Z.shape[0]):
                for j in range(Z.shape[1]):
                    Z[j, i] = ex_fun(np.array([xg[i], yg[j]]))
            return Z

        X, Y = np.meshgrid(xg, yg)
        plt.colorbar(ax.contour(X, Y, mat_fun(xg, yg)))
        fig.show()
        for it, i in zip(res['iterates'], range(len(res['iterates']))):
            ax.plot(it.x[0], it.x[1], '+k')
            ax.annotate(i, it.x)
        plt.show(block=True)

    np.testing.assert_allclose(res.x.reshape(-1),ref.x,atol = 1e-16,rtol = 1e-5)





#
# Copyright 2018 Antoine Sanner
# 
# ### MIT license
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import tests.minimization_problems as mp
import numpy as np
import scipy.optimize

import pytest
@pytest.mark.parametrize("Objective",[mp.Trigonometric,mp.Extended_Rosenbrock])
@pytest.mark.parametrize("n",[2,4,10])
def test_directional_derivative(Objective,n) :
    """
    Asserts the nmerical and the Gradient provided by Objective are close
    :param Objective: which function to minimize
    :param n: number of dimensions
    :return:
    """
    _verbose = False
    if _verbose:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        ax.set_xscale("log")
        ax.set_yscale("log")

    for i in range(12):
        x= Objective.bounds[0] + (Objective.bounds[1] - Objective.bounds[0]) * np.random.random(n)
        x.shape = (-1,1)
        u=np.random.normal(size=n) # Direction
        u /= np.linalg.norm(u,2) # normalize
        u.shape = (-1,1)

        # TODO: Don't know which Tolerance to associate with which eps
        epsilons = np.array([1e-3,1e-5,1e-6])
        #der_numerical = np.zeros_like(epsilons)
        #der_analytical = np.zeros_like(epsilons)
        errorratios = np.zeros_like(epsilons)
        for eps, i in zip(epsilons, range(len(epsilons))):
            #print((Objective.f(x + u * eps) - Objective.f(x)) / eps)
            der_numerical  = (Objective.f(x + u * eps) - Objective.f(x)) / eps
            der_analytical = (Objective.grad(x).T@u).item()

            errorratios[i] =  np.abs(der_numerical - der_analytical) /eps
        errorratios /= errorratios[0]
        if _verbose:
            ax.plot(epsilons,errorratios)

        assert (errorratios < 10).all(), "error_ratios".format(abs(der_numerical- der_analytical)/der_analytical)
    if _verbose:
        plt.show(block=True)
@pytest.mark.parametrize("Objective",[mp.Extended_Rosenbrock,mp.Trigonometric])
@pytest.mark.parametrize("n",[2,4,10,20])
def test_Gradient(Objective,n) :
    """
    Asserts the numerical Gradient converges with order of the step to the analyitical one.

    This uses check_gradient from scipy

    :param Objective: which function to minimize
    :param n: number of dimensions
    :return:
    """
    _verbose = False
    if _verbose:
        import matplotlib.pyplot as plt
        fig,ax = plt.subplots()
        ax.set_xscale("log")
        ax.set_yscale("log")

    for i in range(12): # reproduce to have a bit of statistics
        x= Objective.bounds[0] + (Objective.bounds[1] - Objective.bounds[0]) * np.random.random(n)
        #x.shape = (-1,1)

        epsilons = np.array([1e-3,1e-5,1e-6])
        errors = np.array([scipy.optimize.check_grad(Objective.f,Objective.grad,x,epsilon = eps) for eps in epsilons ]) # should be O(epsilon)
        errorratios = errors / epsilons # should be O(1)
        errorratios /= errorratios[0]
        if _verbose:
            ax.plot(epsilons,errorratios)


        #print(errorratios)
        assert np.prod(errorratios < 10), "errorratios = {}".format(errorratios)
    if _verbose:
        plt.show(block=True)



#
# Copyright 2019 Lars Pastewka
# 
# ### MIT license
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#



import numpy as np

import scipy.optimize

def donothing(*args,**kwargs):
    pass

def getLogfile():
    def savelog(string):
        with open("LBFGS.log", "w") as file:
            file.write("{}".format(string))
    return savelog
    raise BlockingIOError

def steepest_descent_wolfe2(x0,f,fprime,**kwargs):
    """
    For first Iteration there is no history. We make a steepest descent satisfying strong Wolfe Condition
    :return:
    """

    # x_old.shape=(-1,1)
    grad0 = fprime(x0)

    # line search
    alpha, phi, phi0, derphi = scipy.optimize.linesearch.scalar_search_wolfe2(
        lambda alpha: f(x0 - grad0 * alpha),
        lambda alpha: np.dot(fprime(x0 - grad0 * alpha).T, -grad0),**kwargs)
    assert derphi is not None, "Line Search in first steepest descent failed"
    x = x0 - grad0 * alpha

    return x, fprime(x) , x0, grad0



def LBFGS(fun, x, args=(), jac=None, x_old=None, maxcor=5, gtol = None, g2tol=1e-10, maxiter=10000,
          maxls=20, store_iterates="iterate", printdb=donothing, linesearch_options={}):
    """
        This implementation follows

    Byrd, R. H., Nocedal, J. & Schnabel, R. B. REPRESENTATIONS OF QUASI-NEWTON MATRICES AND THEIR USE IN LIMITED MEMORY METHODS. 31

    The signature of the function is such that you can give her as Method in scipy.optimize.minimize

    :param args:
    :param ftol:
    :param maxls:
    :param fun:
    :param jac:
    :param x:
    :param x_old:
    :param maxcor:
    :param g2tol:
    :param maxiter:
    :param store_iterates:
    :param printdb:
    :return:
    """



    if x_old is None:
        x_old = x.copy()

        x,grad,x_old,grad_old = steepest_descent_wolfe2(x_old, fun, jac,**linesearch_options)

    k=1

    n = x.size # Dimension of x

    gamma = 1

    S = np.zeros((n, 0))
    Y = np.zeros((n, 0))
    R = np.zeros((0, 0))
    STgrad = np.array((1, maxcor))
    YTgrad = np.array((1, maxcor))

    grad = np.asarray(jac(x))
    grad2 = np.sum(grad**2)
    grad_old = np.asarray(jac(x_old))

    alpha=0

    # Start loop

    iterates = list()
    if store_iterates == 'iterate':
        iterate = scipy.optimize.OptimizeResult(
            {'x': x_old.copy(),
             'fun': fun(x_old),
             'jac': grad_old.copy()})
        iterates.append(iterate)

        iterate = scipy.optimize.OptimizeResult(
            {'x': x.copy(),
             'fun': fun(x),
             'jac': grad})
        iterates.append(iterate)


    while True:
        #printdb(k)
        #printdb("grads")
        #printdb(grad)
        #printdb(grad_old)

        # Update Sk,Yk
        if k > maxcor:
            S = np.roll(S, -1)
            S[:, -1] = (x - x_old).flat
            Y = np.roll(Y, -1)
            Y[:, -1] = (grad - grad_old).flat

        else:
            S = np.hstack([S, x - x_old])
            Y = np.hstack([Y, grad - grad_old])
        #printdb("S: {}".format(S))
        #printdb("Y: {}".format(Y))

        # 2.
        grad2prev = grad2.copy()
        grad2 = np.sum(grad ** 2)  # ok

        # check if job is done
        if  ((grad2 < g2tol if g2tol is not None else True) and
            (np.max(np.abs(grad)) < gtol if gtol is not None else True)):
            result = scipy.optimize.OptimizeResult({'success': True,
                                                    'x': x,
                                                    'nit': k,
                                                    'iterates': iterates})

            #if iterates:
            #    result['iterates'] = iterates
            return result

        if k > maxiter:
            result = scipy.optimize.OptimizeResult({'success': False,
                                                    'x': x,
                                                    'nit': k,
                                                    'iterates':iterates})

            #if iterates:
            #    result['iterates'] = iterates
            return result

        STgrad_prev = STgrad.copy()
        YTgrad_prev = YTgrad.copy()

        STgrad = np.dot(S.T, grad)
        YTgrad = np.dot(Y.T, grad)  # OK, this is new

        #printdb("STgrad : {}".format(STgrad))
        #printdb("YTgrad: {}".format(YTgrad))

        if k > maxcor:
            w = np.vstack([STgrad_prev, gamma * YTgrad_prev])
            S_now_T_grad_prev = np.roll(STgrad_prev,-1)
            S_now_T_grad_prev[-1] = - alpha * gamma * grad2prev - alpha * w.T.dot(p)
        else : # straightforward Version
            S_now_T_grad_prev = np.dot(S.T, grad_old)

        #printdb("S_now_T_grad_prev {}".format(S_now_T_grad_prev))
        #np.testing.assert_allclose(S_now_T_grad_prev,np.dot(S.T, grad_old),
        #                           err_msg="Maybe the assumption of Theorem 2.2"
        #                                "is not valid: sk-1Tyk-1 = {}".format(S[:,-1].T.dot(Y[:,-1])))

        # 3. # TOOPTIMIZE
        #sprevTgradprev = np.dot(S[:, -1].T, grad_old)  # sk-1T gk-1

        #%% 4.
        #ykm12 = np.dot(Y[:, -1].T, Y[:, -1]) #TOOPTIMIZE

        #printdb("before")
        #printdb("R: {}".format(R))
        if k > maxcor:
            R = np.roll(R, (-1, -1), axis=(0, 1))  # mxm Matrix hold by all Processors
            R[-1, :] = 0
            STym1 = STgrad - S_now_T_grad_prev
            R[:, -1] = STym1.flat  #O(m x n)

        elif k == 1:
            R = np.triu(np.dot(S.T, Y))
        else:
            R = np.vstack([R, np.zeros(k - 1)])
            R = np.hstack([R, np.dot(S.T, Y[:, -1]).reshape(k, 1)])

        #np.testing.assert_allclose(R, np.triu(np.dot(S.T, Y)))

        if k > maxcor:
            D = np.roll(D, (-1, -1), axis=(0, 1))
            # D[-1,-1] = np.dot(Y[:,-1],Y[:,-1])# yk-1Tyk-1 # TOOPTIMIZE
            D[-1, -1] = R[-1,-1]
        else:
            #D = np.diag(np.einsum("ik,ik -> k", S, Y))
            D=np.diag(R.diagonal())

        assert D[-1,-1] >0, "k = {}: ".format(k)  # Assumption of Theorem 2.2
        #np.testing.assert_allclose(np.diag(D), np.diag(R))

        # YTY = np.dot(Y.T,Y) #TOPTIMIZED
        if k > maxcor:
            YTY = np.roll(YTY, (-1, -1), axis=(0, 1))
            #printdb(YTgrad)
            #printdb(YTgrad_prev)
            YTY[-1, :-1] = YTY[:-1, -1] = (YTgrad[:-1] - YTgrad_prev[1:]).flat
            YTY[-1, -1] = grad2prev - grad2 + 2 * YTgrad[-1]
        else:
            YTY = np.dot(Y.T, Y)
        #np.testing.assert_allclose(YTY, np.dot(Y.T, Y))
        ##
        #printdb("after")
        #printdb("R: {}".format(R))
        #printdb("YTY: {}".format(YTY))
        #printdb("D: {}".format(D))

        #%% 5.
        gamma = D[-1, -1] / YTY[-1,-1]  # n.b. D[-1,-1] = sk-1T yk-1 = yk-1T sk-1

        #%% 6.
        #Rinv = np.linalg.inv(R)
        Rinv = scipy.linalg.solve_triangular(R,np.eye(min(k,maxcor)))

        RiSg = Rinv.dot(STgrad)

        p = np.vstack([Rinv.T.dot(D + gamma * YTY).dot(RiSg) - gamma * Rinv.T.dot(YTgrad)
                          , - RiSg])

        #%% 7.
        Hgrad = gamma * grad + np.hstack([S, gamma * Y]).dot(p)

        #%% linesearch

        #reslinesearch = scipy.optimize.minimize_scalar(fun=lambda alpha: fun(x - Hgrad * alpha), bounds=(0, 10), method="bounded")
        #assert reslinesearch.success, "Linesearch not converged"
        # line_search did cause problems, maybe because of the differentz interpretation of the arrays
        #alpha,fc,gc,new_fval,old_fval,new_slope = scipy.optimize.line_search(fun,lambda x_ : fprime(x_).flatten(),x, - Hgrad.flatten() , c1=1e-4,c2=0.9,maxiter=20)

        #printdb("assert descent direction")
        #assert fun(x - Hgrad * 0.001) - fun(x) < 0
        #printdb(fun(x - Hgrad * 0.001) - fun(x))
        
        alpha,phi,phi0,derphi = scipy.optimize.linesearch.scalar_search_wolfe2(lambda alpha: fun(x - Hgrad * alpha), lambda alpha: np.dot(jac(x - Hgrad * alpha).T, -Hgrad),maxiter = maxls,**linesearch_options)

        if derphi is None:
            import matplotlib.pyplot as plt
            figdebug,axdebug=plt.subplots()
            alphas = np.linspace(-1,10)

            axdebug.plot(alphas,[fun(x - a * Hgrad) for a in alphas] )
            figdebug.show()
            printdb("scalar line search did not converge")
            printdb("alpha: {}".format(alpha))
            plt.show(block=True)

        assert derphi is not None, "scalar line-search did not converge"
        #assert new_fval is not None, "Line-search didn't converge"
        #printdb("x: {}".format(x))
        x_old[:] = x
        #printdb("x_old: {}".format(x_old))
        x = x - Hgrad * alpha
        #printdb("x: {}".format(x))
        #printdb("x_old: {}".format(x_old))
        #printdb("x = {}".format(x))
        #assert phi < phi0, "f(x) >= f(x_old) ! "
        grad_old[:] = grad
        grad = jac(x)
        #assert fun(x) <= fun(x_old) + 1e-4 * alpha * grad_old.T.dot(-Hgrad), "First Wolfe Condition not fullfilled"
        #assert grad.T.dot(-Hgrad) >= 0.9 * grad_old.T.dot(-Hgrad), "second Wolfe Condition not fullfilled"
        #printdb("dx * -Hgrad:{}".format((x-x_old).T.dot(-Hgrad)))
        #printdb(alpha)
        #assert (grad - grad_old).T.dot(x - x_old) > 0, "msg derphi = {}".format(derphi)
        if store_iterates == 'iterate':
            iterate = scipy.optimize.OptimizeResult(
                {'x': x.copy(),
                 'fun': phi,
                 'jac': grad})
            iterates.append(iterate)

        k = k + 1

#alpha = np.linspace(0, 2)
#plt.plot(alpha, [test_fun(x_old - grad_old * a) for a in alpha])
#plt.axvline(step)